# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
import pyecvl._core.ecvl as ecvl_core
import pyecvl.ecvl as ecvl_py


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_size(ecvl):
    assert ecvl.DataTypeSize(ecvl.DataType.int8) == 1
    assert ecvl.DataTypeSize(ecvl.DataType.int16) == 2
    assert ecvl.DataTypeSize(ecvl.DataType.int32) == 4
    assert ecvl.DataTypeSize(ecvl.DataType.int64) == 8
    assert ecvl.DataTypeSize(ecvl.DataType.float32) == 4
    assert ecvl.DataTypeSize(ecvl.DataType.float64) == 8
    assert ecvl.DataTypeSize(ecvl.DataType.uint8) == 1
    assert ecvl.DataTypeSize(ecvl.DataType.uint16) == 2


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_size_no_args(ecvl):
    assert ecvl.DataTypeSize() == 8
    assert ecvl.DataTypeSignedSize() == 6



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import pytest

import pyecvl._core.ecvl as ecvl_core
import pyecvl.ecvl as ecvl_py


def _empty_img(ecvl):
    if ecvl is ecvl_core:
        return ecvl.Image()
    return ecvl.Image.empty()


def _img_fromarray(ecvl, array, channels, colortype, spacings=None):
    if spacings is None:
        spacings = []
    if ecvl is ecvl_core:
        return ecvl.Image(array, channels, colortype, spacings)
    return ecvl.Image.fromarray(array, channels, colortype, spacings)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_empty(ecvl):
    img = _empty_img(ecvl)
    assert len(img.dims_) == 0
    assert img.IsEmpty()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_copy(ecvl):
    dims, dtype = [2, 4, 3], ecvl.DataType.int8
    ch, ctype = "xyc", ecvl.ColorType.RGB
    x = ecvl.Image(dims, dtype, ch, ctype)
    a = np.array(x, copy=False)
    a.fill(3)
    y = x.copy()
    assert y.dims_ == dims
    assert y.elemtype_ == dtype
    assert y.channels_ == ch
    assert y.colortype_ == ctype
    b = np.array(y, copy=False)
    assert (b == 3).all()
    # ecvl_core can also use copy constructor binding
    if ecvl is ecvl_core:
        y = ecvl.Image(x)
        assert y.dims_ == dims
        assert y.elemtype_ == dtype
        assert y.channels_ == ch
        assert y.colortype_ == ctype
        b = np.array(y, copy=False)
        assert (b == 3).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_five_dims(ecvl):
    dims = [1, 2, 3, 4, 5]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyzoo", ecvl.ColorType.none)
    assert img.dims_ == dims
    assert len(img.strides_) == len(dims)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_height_width_channels(ecvl):
    img = ecvl.Image([3, 10, 8], ecvl.DataType.int8, "cxy", ecvl.ColorType.BGR)
    assert img.Channels() == 3
    assert img.Width() == 10
    assert img.Height() == 8


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_extra_args(ecvl):
    dims = [1, 2, 3, 4, 5]
    spacings = [1.1, 1.1]
    dev = ecvl.Device.CPU
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyzoo", ecvl.ColorType.none,
                     spacings, dev)
    assert img.dims_ == dims
    assert len(img.strides_) == len(dims)
    assert img.spacings_ == pytest.approx(spacings)
    assert img.dev_ == dev
    img.To(ecvl.Device.CPU)
    assert img.dev_ == dev


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_view(ecvl):
    x = ecvl.Image([5, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    assert x.Channels() == 3
    assert x.IsOwner()
    y = ecvl.View_int8(x)
    assert not y.IsOwner()
    y[1, 2, 0] = 36
    y[3, 3, 2] = 48
    y[4, 2, 1] = -127
    y[3, 2, 0] = -128
    dst = _empty_img(ecvl)
    ecvl.Neg(x, dst, ecvl.DataType.int8, False)
    dst_view = ecvl.View_int8(dst)
    assert dst_view[1, 2, 0] == -36
    assert dst_view[3, 3, 2] == -48
    assert dst_view[4, 2, 1] == 127
    assert dst_view[3, 2, 0] == -128


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_rearrange_channels(ecvl):
    S = [3, 4, 3, 2]
    img = ecvl.Image(S, ecvl.DataType.int16, "cxyz", ecvl.ColorType.RGB)
    view = ecvl.View_int16(img)
    for i in range(S[0]):
        for j in range(S[1]):
            for k in range(S[2]):
                for l in range(S[3]):  # noqa
                    view[i, j, k, l] = (l + k * S[3] + j * S[2] * S[3] +
                                        i * S[1] * S[2] * S[3])
    img2 = _empty_img(ecvl)
    ecvl.RearrangeChannels(img, img2, "xyzc")
    view2 = ecvl.View_int16(img2)
    assert view2[2, 0, 1, 0] == view[0, 2, 0, 1]
    assert view2[3, 1, 1, 2] == view[2, 3, 1, 1]
    assert view2[0, 2, 0, 1] == view[1, 0, 2, 0]
    assert view2[1, 2, 0, 1] == view[1, 1, 2, 0]


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_numpy(ecvl):
    shape = [2, 3, 4]
    img = ecvl.Image(shape, ecvl.DataType.int16, "xyc", ecvl.ColorType.none)
    view = ecvl.View_int16(img)
    a = np.array(img, copy=False)
    assert list(a.shape) == shape
    assert list(a.strides) == img.strides_
    assert a.dtype == np.int16
    for i in range(shape[0]):
        for j in range(shape[1]):
            for k in range(shape[2]):
                assert a[i, j, k] == view[i, j, k]
    a.flat = 42
    b = np.array(img, copy=True)
    assert np.array_equal(a + img, a + b)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_numpy_types(ecvl):
    shape, channels, color_type = [2, 3], "xy", ecvl.ColorType.none
    for dt_name in ("int8", "int16", "int32", "int64",
                    "float32", "float64", "uint8", "uint16"):
        np_dt = getattr(np, dt_name)
        dt = getattr(ecvl.DataType, dt_name)
        img = ecvl.Image(shape, dt, channels, color_type)
        a = np.array(img, copy=False)
        assert a.dtype == np_dt


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_image_from_array(ecvl):
    channels, color_type = "xy", ecvl.ColorType.none
    for dt_name in ("int8", "int16", "int32", "int64",
                    "float32", "float64", "uint8", "uint16"):
        np_dt = getattr(np, dt_name)
        dt = getattr(ecvl.DataType, dt_name)
        a = np.arange(12).reshape(3, 4).astype(np_dt)
        b = np.asfortranarray(a)
        img = _img_fromarray(ecvl, b, channels, color_type)
        assert img.elemtype_ == dt
        assert img.dims_ == list(b.shape)
        assert img.strides_ == list(b.strides)
        # check construction from c-style array
        img = _img_fromarray(ecvl, a, channels, color_type)
        assert img.elemtype_ == dt
        assert img.dims_ == list(a.shape)
        assert img.strides_ == list(b.strides)
    a = np.arange(12).reshape(3, 4).astype(np.int16)
    b = np.asfortranarray(a)
    img = _img_fromarray(ecvl, b, channels, color_type)
    view = ecvl.View_int16(img)
    for i in range(3):
        for j in range(4):
            assert view[i, j] == b[i, j]
    # check construction from c-style array
    img = _img_fromarray(ecvl, a, channels, color_type)
    view = ecvl.View_int16(img)
    for i in range(3):
        for j in range(4):
            assert view[i, j] == a[i, j]


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_add(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    y = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    b = np.array(y, copy=False)
    a.fill(10)
    b.fill(3)
    x.Add(y)
    assert (a == 13).all()
    a.fill(10)
    x += y
    assert (a == 13).all()
    a.fill(10)
    b.fill(118)
    x.Add(y)
    assert (a == 127).all()
    a.fill(10)
    x.Add(y, saturate=False)
    assert (a == -128).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_sub(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    y = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    b = np.array(y, copy=False)
    a.fill(-10)
    b.fill(3)
    x.Sub(y)
    assert (a == -13).all()
    a.fill(-10)
    x -= y
    assert (a == -13).all()
    a.fill(-10)
    b.fill(119)
    x.Sub(y)
    assert (a == -128).all()
    a.fill(-10)
    x.Sub(y, saturate=False)
    assert (a == 127).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_mul(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    y = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    b = np.array(y, copy=False)
    a.fill(16)
    b.fill(3)
    x.Mul(y)
    assert (a == 48).all()
    a.fill(16)
    x *= y
    assert (a == 48).all()
    a.fill(16)
    b.fill(8)
    x.Mul(y)
    assert (a == 127).all()
    a.fill(16)
    x.Mul(y, saturate=False)
    assert (a == -128).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_div(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    y = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    b = np.array(y, copy=False)
    a.fill(11)
    b.fill(2)
    x.Div(y)
    assert (a == 5).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_copy_image(ecvl):
    dims, dtype = [2, 4, 3], ecvl.DataType.int8
    ch, ctype = "xyc", ecvl.ColorType.RGB
    x = ecvl.Image(dims, dtype, ch, ctype)
    a = np.array(x, copy=False)
    a.fill(3)
    # without dtype arg
    y = _empty_img(ecvl)
    ecvl.CopyImage(x, y)
    assert y.dims_ == dims
    assert y.elemtype_ == dtype
    assert y.channels_ == ch
    assert y.colortype_ == ctype
    b = np.array(y, copy=False)
    assert (b == 3).all()
    # with dtype arg
    new_dtype = ecvl.DataType.float32
    y = _empty_img(ecvl)
    ecvl.CopyImage(x, y, new_dtype)
    assert y.dims_ == dims
    assert y.elemtype_ == new_dtype
    assert y.channels_ == ch
    assert y.colortype_ == ctype
    b = np.array(y, copy=False)
    assert (b == 3).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_shallow_copy_image(ecvl):
    dims, dtype = [2, 4, 3], ecvl.DataType.int8
    ch, ctype = "xyc", ecvl.ColorType.RGB
    x = ecvl.Image(dims, dtype, ch, ctype)
    a = np.array(x, copy=False)
    a.fill(3)
    y = _empty_img(ecvl)
    ecvl.ShallowCopyImage(x, y)
    assert y.dims_ == dims
    assert y.elemtype_ == dtype
    assert y.channels_ == ch
    assert y.colortype_ == ctype
    b = np.array(y, copy=False)
    assert (b == 3).all()
    b.fill(4)
    assert (a == 4).all()
    assert x.IsOwner()
    assert not y.IsOwner()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ConvertTo(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.RGB)
    y = _empty_img(ecvl)
    a = np.array(x, copy=False)
    a.fill(128)
    # saturate = False
    ecvl.ConvertTo(x, y, ecvl.DataType.int8, False)
    assert y.elemtype_ == ecvl.DataType.int8
    b = np.array(y, copy=False)
    assert (b == -128).all()
    # saturate = True
    ecvl.ConvertTo(x, y, ecvl.DataType.int8, True)
    assert y.elemtype_ == ecvl.DataType.int8
    b = np.array(y, copy=False)
    assert (b == 127).all()
    # saturate = True by default
    ecvl.ConvertTo(x, y, ecvl.DataType.int8)
    assert y.elemtype_ == ecvl.DataType.int8
    b = np.array(y, copy=False)
    assert (b == 127).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ConvertTo_method(ecvl):
    # saturate = False
    x = ecvl.Image([2, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    a.fill(128)
    x.ConvertTo(ecvl.DataType.int8, False)
    assert x.elemtype_ == ecvl.DataType.int8
    a = np.array(x, copy=False)
    assert (a == -128).all()
    # saturate = True
    x = ecvl.Image([2, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    a.fill(128)
    x.ConvertTo(ecvl.DataType.int8, True)
    assert x.elemtype_ == ecvl.DataType.int8
    a = np.array(x, copy=False)
    assert (a == 127).all()
    # saturate = True by default
    x = ecvl.Image([2, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    a.fill(128)
    x.ConvertTo(ecvl.DataType.int8)
    assert x.elemtype_ == ecvl.DataType.int8
    a = np.array(x, copy=False)
    assert (a == 127).all()



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .imgproc import InterpolationType, BorderType
from . import _core
_ecvl = _core.ecvl


if not _ecvl.ECVL_EDDL:
    raise ImportError("extension module not compiled with EDDL support")


__all__ = [
    "AugmentationParam",
    "AugmentationFactory",
    "SequentialAugmentationContainer",
    "OneOfAugmentationContainer",
    "AugRotate",
    "AugResizeDim",
    "AugResizeScale",
    "AugFlip",
    "AugMirror",
    "AugGaussianBlur",
    "AugAdditiveLaplaceNoise",
    "AugAdditivePoissonNoise",
    "AugGammaContrast",
    "AugCoarseDropout",
    "AugTranspose",
    "AugBrightness",
    "AugGridDistortion",
    "AugElasticTransform",
    "AugOpticalDistortion",
    "AugSalt",
    "AugPepper",
    "AugSaltAndPepper",
    "AugNormalize",
    "AugCenterCrop",
    "AugToFloat32",
    "AugDivBy255",
    "AugScaleTo",
]


class AugmentationParam(_ecvl.AugmentationParam):
    r"""\
    Augmentation parameters which must be randomly generated within a range.

    :var min\_: minimum value for the random range
    :var max\_: maximum value for the random range
    :var value\_: generated parameter value
    """
    def GenerateValue(self):
        r"""\
        Generate the random value between min\_ and max\_.
        """
        return _ecvl.AugmentationParam.GenerateValue(self)

    @staticmethod
    def SetSeed(seed):
        """\
        Set a fixed seed for the random generated values.

        Useful to reproduce experiments with same augmentations.

        :param seed: seed value
        :return: None
        """
        return _ecvl.AugmentationParam.SetSeed(seed)


class AugmentationFactory(_ecvl.AugmentationFactory):
    r"""\
    Creates augmentations from text strings.

    If only one argument is supplied, it needs to include the augmentation's
    name, e.g.::

        AugmentationFactory.create('AugFlip p=0.5')

    If only two arguments are supplied, the first is the augmentation's
    name, e.g.::

        AugmentationFactory.create('AugFlip', 'p=0.5')
    """

    @staticmethod
    def create(s1, s2=None):
        if s2 is None:
            return _ecvl.AugmentationFactory.create(s1)
        return _ecvl.AugmentationFactory.create(s1, s2)


class SequentialAugmentationContainer(_ecvl.SequentialAugmentationContainer):
    """\
    A container for multiple augmentations to be applied in sequence.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create a SequentialAugmentationContainer from a text description,
        e.g.::

            txt = '''\
            AugFlip p=0.2
            AugMirror p=0.4
            end
            '''
            c = SequentialAugmentationContainer(txt)
        """
        return _ecvl.SequentialAugmentationContainer(txt)

    def __init__(self, augs):
        """\
        :param augs: list of augmentations to be applied
        """
        _ecvl.SequentialAugmentationContainer.__init__(self, augs)


class OneOfAugmentationContainer(_ecvl.OneOfAugmentationContainer):
    """\
    A container for multiple augmentations, from which one is randomly chosen.

    The chosen augmentation will be applied with a user-specified probability.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create a OneOfAugmentationContainer from a text description,
        e.g.::

            txt = '''\
            p=0.7
            AugFlip p=0.2
            AugMirror p=0.4
            end
            '''
            c = OneOfAugmentationContainer(txt)
        """
        return _ecvl.OneOfAugmentationContainer(txt)

    def __init__(self, p, augs):
        """\
        :param p: probability of applying the chosen augmentation
        :param augs: list of augmentations
        """
        _ecvl.OneOfAugmentationContainer.__init__(self, p, augs)


class AugRotate(_ecvl.AugRotate):
    """\
    Augmentation wrapper for Rotate2D.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugRotate from a text description, e.g.::

            a = AugRotate('angle=[30, 50] center=(2, 3) '
                          'scale=1.1 interp="nearest"')
        """
        return _ecvl.AugRotate(txt)

    def __init__(self, angle, center=None, scale=1.0,
                 interp=InterpolationType.linear,
                 gt_interp=InterpolationType.nearest):
        """\
        :param angle: range of degrees ``[min, max]`` to randomly select from
        :param center: a list of floats representing the coordinates of the
          rotation center. If None, the center of the image is used
        :param scale: scaling factor
        :param interp: InterpolationType to be used
        :param gt_interp: InterpolationType to be used for ground truth
        """
        if center is None:
            center = []
        _ecvl.AugRotate.__init__(self, angle, center, scale, interp)


class AugResizeDim(_ecvl.AugResizeDim):
    """\
    Augmentation wrapper for ResizeDim.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugResizeDim from a text description, e.g.::

            a = AugResizeDim('dims=(4, 3) interp="linear"')
        """
        return _ecvl.AugResizeDim(txt)

    def __init__(self, dims, interp=InterpolationType.linear,
                 gt_interp=InterpolationType.nearest):
        """\
        :param dims: list of integers that specifies the new size of each
          dimension
        :param interp: InterpolationType to be used
        :param gt_interp: InterpolationType to be used for ground truth
        """
        _ecvl.AugResizeDim.__init__(self, dims, interp)


class AugResizeScale(_ecvl.AugResizeScale):
    """\
    Augmentation wrapper for ResizeScale.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugResizeScale from a text description, e.g.::

            a = AugResizeScale('scale=(0.5, 0.5) interp="linear"')
        """
        return _ecvl.AugResizeScale(txt)

    def __init__(self, scale, interp=InterpolationType.linear,
                 gt_interp=InterpolationType.nearest):
        """\
        :param scale: list of floats that specifies the scale to apply to
          each dimension
        :param interp: InterpolationType to be used
        :param gt_interp: InterpolationType to be used for ground truth
        """
        _ecvl.AugResizeScale.__init__(self, scale, interp)


class AugFlip(_ecvl.AugFlip):
    """\
    Augmentation wrapper for Flip2D.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugFlip from a text description, e.g.::

            a = AugFlip('p=0.5')
        """
        return _ecvl.AugFlip(txt)

    def __init__(self, p=0.5):
        """\
        :param p: probability of each image to get flipped
        """
        _ecvl.AugFlip.__init__(self, p)


class AugMirror(_ecvl.AugMirror):
    """\
    Augmentation wrapper for Mirror2D.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugMirror from a text description, e.g.::

            a = AugMirror('p=0.5')
        """
        return _ecvl.AugMirror(txt)

    def __init__(self, p=0.5):
        """\
        :param p: probability of each image to get mirrored
        """
        _ecvl.AugMirror.__init__(self, p)


class AugGaussianBlur(_ecvl.AugGaussianBlur):
    """\
    Augmentation wrapper for GaussianBlur.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugGaussianBlur from a text description, e.g.::

            a = AugGaussianBlur('sigma=[0.2, 0.4]')
        """
        return _ecvl.AugGaussianBlur(txt)

    def __init__(self, sigma):
        """\
        :param sigma: sigma range ``[min, max]`` to randomly select from.
        """
        _ecvl.AugGaussianBlur.__init__(self, sigma)


class AugAdditiveLaplaceNoise(_ecvl.AugAdditiveLaplaceNoise):
    """\
    Augmentation wrapper for AdditiveLaplaceNoise.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugAdditiveLaplaceNoise from a text description, e.g.::

            a = AugAdditiveLaplaceNoise('std_dev=[12.5, 23.1]')
        """
        return _ecvl.AugAdditiveLaplaceNoise(txt)

    def __init__(self, std_dev):
        """\
        :param std_dev: range of values ``[min, max]`` to randomly select the \
          standard deviation of the noise generating distribution. Suggested \
          values are around 255 * 0.05 for uint8 images
        """
        _ecvl.AugAdditiveLaplaceNoise.__init__(self, std_dev)


class AugAdditivePoissonNoise(_ecvl.AugAdditivePoissonNoise):
    """\
    Augmentation wrapper for AdditivePoissonNoise.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugAdditivePoissonNoise from a text description, e.g.::

            a = AugAdditivePoissonNoise('lambda=[2.0, 3.0]')
        """
        return _ecvl.AugAdditivePoissonNoise(txt)

    def __init__(self, lambda_):
        """\
        :param lambda\_: range of values ``[min, max]`` to randomly select \
          the lambda of the noise generating distribution. Suggested values \
          are around 0.0 to 10.0.
        """  # noqa
        _ecvl.AugAdditivePoissonNoise.__init__(self, lambda_)


class AugGammaContrast(_ecvl.AugGammaContrast):
    """\
    Augmentation wrapper for GammaContrast.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugGammaContrast from a text description, e.g.::

            a = AugGammaContrast('gamma=[3, 4]')
        """
        return _ecvl.AugGammaContrast(txt)

    def __init__(self, gamma):
        """\
        :param gamma: range of values ``[min, max]`` to randomly select the \
          exponent for the contrast adjustment. Suggested values are around \
          0.5 to 2.0.
        """
        _ecvl.AugGammaContrast.__init__(self, gamma)


class AugCoarseDropout(_ecvl.AugCoarseDropout):
    """\
    Augmentation wrapper for CoarseDropout.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugCoarseDropout from a text description, e.g.::

            a = AugCoarseDropout('p=[0.5, 0.7] drop_size=[0.1, 0.2] '
                                 'per_channel=0.4')
        """
        return _ecvl.AugCoarseDropout(txt)

    def __init__(self, p, drop_size, per_channel):
        """\
        :param p: range of values ``[min, max]`` to randomly select the
          probability of any rectangle being set to zero
        :param drop_size: range of values ``[min, max]`` to randomly select
          the size of rectangles in percentage of the input image
        :param per_channel: probability of each image to use the same value
          for all channels of a pixel
        """
        _ecvl.AugCoarseDropout.__init__(self, p, drop_size, per_channel)


class AugTranspose(_ecvl.AugTranspose):
    """\
    Augmentation wrapper for Transpose.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugTranspose from a text description, e.g.::

            a = AugTranspose('p=0.5')
        """
        return _ecvl.AugTranspose(txt)

    def __init__(self, p=0.5):
        """\
        :param p: probability of each image to get transposed.
        """
        _ecvl.AugTranspose.__init__(self, p)


class AugBrightness(_ecvl.AugBrightness):
    """\
    Augmentation wrapper for brightness adjustment.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugBrightness from a text description, e.g.::

            a = AugBrightness('beta=[30, 60]')
        """
        return _ecvl.AugBrightness(txt)

    def __init__(self, beta):
        """\
        :param beta: range of values ``[min, max]`` to randomly select from \
          for the brightness adjustment. Suggested values are around 0 to 100.
        """
        _ecvl.AugBrightness.__init__(self, beta)


class AugGridDistortion(_ecvl.AugGridDistortion):
    """\
    Augmentation wrapper for GridDistortion.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugGridDistortion from a text description, e.g.::

            a = AugGridDistortion('num_steps=[5,10] distort_limit=[-0.2,0.2] '
                                  'interp=\"linear\" '
                                  'border_type=\"reflect_101\" '
                                  'border_value=0')
        """
        return _ecvl.AugGridDistortion(txt)

    def __init__(self, num_steps, distort_limit,
                 interp=InterpolationType.linear,
                 border_type=BorderType.BORDER_REFLECT_101, border_value=0):
        """\
        :param num_steps: range of values ``[min, max]`` to randomly select
          the number of grid cells on each side
        :param distort_limit: range of values ``[min, max]`` to randomly select
          the distortion steps
        :param interp: InterpolationType to be used
        :param border_type: pixel extrapolation method, see BorderType
        :param border_value: padding value if border_type is
          BorderType.BORDER_CONSTANT
        """
        _ecvl.AugGridDistortion.__init__(self, num_steps, distort_limit,
                                         interp, border_type, border_value)


class AugElasticTransform(_ecvl.AugElasticTransform):
    """\
    Augmentation wrapper for ElasticTransform.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugElasticTransform from a text description, e.g.::

            a = AugElasticTransform('alpha=[34,60] sigma=[4,6] '
                                    'interp=\"linear\" '
                                    'border_type=\"reflect_101\" '
                                    'border_value=0')
        """
        return _ecvl.AugElasticTransform(txt)

    def __init__(self, alpha, sigma, interp=InterpolationType.linear,
                 border_type=BorderType.BORDER_REFLECT_101, border_value=0):
        """\
        :param alpha: range of values ``[min, max]`` to randomly select the
          scaling factor that controls the intensity of the deformation
        :param sigma: range of values ``[min, max]`` to randomly select the
          gaussian kernel standard deviation
        :param interp: InterpolationType to be used
        :param border_type: pixel extrapolation method, see BorderType
        :param border_value: padding value if border_type is
          BorderType.BORDER_CONSTANT
        """
        _ecvl.AugElasticTransform.__init__(self, alpha, sigma, interp,
                                           border_type, border_value)


class AugOpticalDistortion(_ecvl.AugOpticalDistortion):
    """\
    Augmentation wrapper for OpticalDistortion.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugOpticalDistortion from a text description, e.g.::

            a = AugOpticalDistortion('distort_limit=[-0.2,0.2] '
                                     'shift_limit=[-0.4,0.4] '
                                     'interp=\"linear\" '
                                     'border_type=\"reflect_101\" '
                                     'border_value=0')
        """
        return _ecvl.AugOpticalDistortion(txt)

    def __init__(self, distort_limit, shift_limit,
                 interp=InterpolationType.linear,
                 border_type=BorderType.BORDER_REFLECT_101, border_value=0):
        """\
        :param distort_limit: range of values ``[min, max]`` to randomly select
          the distortion steps
        :param shift_limit: range of values ``[min, max]`` to randomly select
          the image shifting
        :param interp: InterpolationType to be used
        :param border_type: pixel extrapolation method, see BorderType
        :param border_value: padding value if border_type is
          BorderType.BORDER_CONSTANT
        """
        _ecvl.AugOpticalDistortion.__init__(self, distort_limit, shift_limit,
                                            interp, border_type, border_value)


class AugSalt(_ecvl.AugSalt):
    """\
    Augmentation wrapper for Salt.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugSalt from a text description, e.g.::

            a = AugSalt('p=[0.1,0.3] per_channel=0.5')
        """
        return _ecvl.AugSalt(txt)

    def __init__(self, p, per_channel):
        """\
        :param p: range of values ``[min, max]`` for the probability of any
          pixel to be set to white
        :param per_channel: probability to use the same value for all channels
        """
        _ecvl.AugSalt.__init__(self, p, per_channel)


class AugPepper(_ecvl.AugPepper):
    """\
    Augmentation wrapper for Pepper.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugPepper from a text description, e.g.::

            a = AugPepper('p=[0.1,0.3] per_channel=0.5')
        """
        return _ecvl.AugPepper(txt)

    def __init__(self, p, per_channel):
        """\
        :param p: range of values ``[min, max]`` for the probability of any
          pixel to be set to black
        :param per_channel: probability to use the same value for all channels
        """
        _ecvl.AugPepper.__init__(self, p, per_channel)


class AugSaltAndPepper(_ecvl.AugSaltAndPepper):
    """\
    Augmentation wrapper for SaltAndPepper.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugSaltAndPepper from a text description, e.g.::

            a = AugSaltAndPepper('p=[0.1,0.3] per_channel=0.5')
        """
        return _ecvl.AugSaltAndPepper(txt)

    def __init__(self, p, per_channel):
        """\
        :param p: range of values ``[min, max]`` for the probability of any
          pixel to be set to white or black
        :param per_channel: probability to use the same value for all channels
        """
        _ecvl.AugSaltAndPepper.__init__(self, p, per_channel)


class AugNormalize(_ecvl.AugNormalize):
    """\
    Augmentation wrapper for Normalize.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugNormalize from a text description, e.g.::

            a = AugNormalize('mean=20 std=5.5')

        Separate mean and std for xyc images::

            a = AugNormalize('mean=(20,19,21) std=(5,5.5,6)')
        """
        return _ecvl.AugNormalize(txt)

    def __init__(self, mean, std):
        """\
        For xyc input images, ``mean`` and ``std`` can be lists of floats,
        representing separate mean and standard deviation for each color
        channel.

        :param mean: mean to use for normalization
        :param std: standard deviation to use for normalization
        """
        _ecvl.AugNormalize.__init__(self, mean, std)


class AugCenterCrop(_ecvl.AugCenterCrop):
    """\
    Augmentation wrapper for CenterCrop.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugCenterCrop from a text description, e.g.::

            a = AugCenterCrop('size=(10, 20)')
        """
        return _ecvl.AugCenterCrop(txt)

    def __init__(self, size):
        """\
        :param size: list of integers [w, h] specifying the output size
        """
        _ecvl.AugCenterCrop.__init__(self, size)


class AugToFloat32(_ecvl.AugToFloat32):
    """\
    Augmentation ToFloat32.

    Converts an Image (and ground truth) to DataType::float32 dividing it by
    divisor (or divisor_gt) parameter.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugToFloat32 from a text description, e.g.::

            a = AugToFloat32('divisor=2. divisor_gt=3.')
        """
        return _ecvl.AugToFloat32(txt)

    def __init__(self, divisor=1., divisor_gt=1.):
        """\
        :param divisor: divisor for the image
        :param divisor: divisor for the ground truth
        """
        _ecvl.AugToFloat32.__init__(self, divisor, divisor_gt)


class AugDivBy255(_ecvl.AugDivBy255):
    """\
    Augmentation DivBy255.

    Divides an Image (and ground truth) by 255.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugDivBy255 from a text description, e.g.::

            a = AugDivBy255('')
        """
        return _ecvl.AugDivBy255(txt)

    def __init__(self):
        """\
        """
        _ecvl.AugDivBy255.__init__(self)


class AugScaleTo(_ecvl.AugScaleTo):
    """\
    Augmentation wrapper for ScaleTo.
    """

    @staticmethod
    def fromtext(txt):
        r"""\
        Create an AugScaleTo from a text description, e.g.::

            a = AugScaleTo('new_min=1 new_max=255')
        """
        return _ecvl.AugScaleTo(txt)

    def __init__(self, new_min, new_max):
        """\
        :param new_min: new minimum value
        :param new_max: new maximum value
        """
        _ecvl.AugScaleTo.__init__(self, new_min, new_max)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
ECVL-EDDL interaction.
"""

import argparse
import sys

import pyecvl.ecvl as ecvl
try:
    from pyeddl.tensor import Tensor
except ImportError:
    print("PyEDDL not installed - quitting")
    sys.exit(0)


AUG_TXT = '''\
SequentialAugmentationContainer
    AugRotate angle=[-5,5] center=(0,0) scale=0.5 interp="linear"
    AugAdditiveLaplaceNoise std_dev=[0,0.51]
    AugCoarseDropout p=[0,0.55] drop_size=[0.02,0.1] per_channel=0
    AugAdditivePoissonNoise lambda=[0,40]
    AugResizeDim dims=(30,30) interp="linear"
end
'''


def main(args):
    if not ecvl.ECVL_EDDL:
        print("No EDDL support - quitting")
        sys.exit(0)
    img = ecvl.ImRead(args.in_img)
    augs = ecvl.SequentialAugmentationContainer([
        ecvl.AugRotate([-5, 5]),
        ecvl.AugMirror(.5),
        ecvl.AugFlip(.5),
        ecvl.AugGammaContrast([3, 5]),
        ecvl.AugAdditiveLaplaceNoise([0, 0.2 * 255]),
        ecvl.AugCoarseDropout([0, 0.55], [0.02, 0.1], 0.5),
        ecvl.AugAdditivePoissonNoise([0, 40]),
        ecvl.AugResizeDim([500, 500]),
    ])
    ecvl.AugmentationParam.SetSeed(0)
    augs.Apply(img)
    print("Executing ImageToTensor")
    t = ecvl.ImageToTensor(img)
    t.div_(128)
    t.mult_(128)
    print("Executing TensorToImage")
    img = ecvl.TensorToImage(t)
    print("Executing TensorToView")
    ecvl.TensorToView(t)

    _ = ecvl.AugmentationFactory.create(AUG_TXT)

    training_augs = ecvl.SequentialAugmentationContainer([
        ecvl.AugRotate([-5, 5]),
        ecvl.AugAdditiveLaplaceNoise([0, 0.2 * 255]),
        ecvl.AugCoarseDropout([0, 0.55], [0.02, 0.1], 0),
        ecvl.AugAdditivePoissonNoise([0, 40]),
        ecvl.AugResizeDim([30, 30]),
    ])
    test_augs = ecvl.SequentialAugmentationContainer([
        ecvl.AugResizeDim([30, 30]),
    ])
    ds_augs = ecvl.DatasetAugmentations([training_augs, None, test_augs])

    batch_size = 64
    print("Creating a DLDataset")
    d = ecvl.DLDataset(args.in_ds, batch_size, ds_augs, ecvl.ColorType.GRAY)
    print("Create x and y")
    x = Tensor(
        [batch_size, d.n_channels_, d.resize_dims_[0], d.resize_dims_[1]]
    )
    y = Tensor([batch_size, len(d.classes_)])

    # Load a batch of d.batch_size_ images into x and corresponding labels
    # into y. Images are resized to the dimensions specified in the
    # augmentations chain
    print("Executing LoadBatch on training set")
    d.LoadBatch(x, y)

    # Change colortype and channels
    img = ecvl.TensorToImage(x)
    img.colortype_ = ecvl.ColorType.GRAY
    img.channels_ = "xyc"

    # Switch to Test split and load a batch of images
    print("Executing LoadBatch on test set")
    d.SetSplit(ecvl.SplitType.test)
    d.LoadBatch(x, y)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("in_img", metavar="INPUT_IMAGE")
    parser.add_argument("in_ds", metavar="INPUT_DATASET")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
DeepHealth dataset generation.
"""

import argparse
import os
import sys

import pyecvl.ecvl as ecvl


def main(args):

    # Classification dataset
    gcd = ecvl.GenerateClassificationDataset(args.cls_dir)
    cls_d = gcd.GetDataset()
    print("dumping classification dataset")
    cls_d.Dump("classification_dataset.yml")

    # Segmentation dataset
    # Possible ground truth suffix or extension if different from images
    suffix = "_segmentation.png"
    # Possible ground truth name for images that have the same ground truth
    gt_name = "black.png"
    gsd = ecvl.GenerateSegmentationDataset(args.seg_dir, suffix, gt_name)
    seg_d = gsd.GetDataset()
    # remove from training set all the images with a "black" ground truth
    seg_d.split_.training_ = [
        i for i in seg_d.split_.training_
        if os.path.basename(seg_d.samples_[i].label_path_) != "black.png"
    ]
    print("dumping segmentation dataset")
    seg_d.Dump("segmentation_dataset.yml")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("cls_dir", metavar="CLASSIFICATION_INPUT_DIR")
    parser.add_argument("seg_dir", metavar="SEGMENTATION_INPUT_DIR")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .datatype import DataType
from . import _core
_ecvl = _core.ecvl


__all__ = [
    "Neg",
    "Add",
    "Sub",
    "Mul",
    "Div",
]


def Neg(src, dst, dst_type=DataType.none, saturate=True):
    """\
    In-place negation of an image.

    Negates every value of ``src``, and stores the the result in ``dst``
    with the specified type.

    :param src: source image
    :param dst: destination image
    :param dst_type: destination image DataType
    :param saturate: whether to apply saturation
    :return: None
    """
    return _ecvl.Neg(src, dst, dst_type, saturate)


def Add(src1, src2, dst, dst_type=DataType.none, saturate=True):
    """\
    Add two images.

    Adds ``src1`` to ``src2`` and stores the the result in ``dst`` with the
    specified type.

    :param src1: source image 1
    :param src2: source image 2
    :param dst: destination image
    :param dst_type: destination image DataType
    :param saturate: whether to apply saturation
    :return: None

    """
    return _ecvl.Add(src1, src2, dst, dst_type, saturate)


def Sub(src1, src2, dst, dst_type=DataType.none, saturate=True):
    """\
    Subtract an image from another.

    Subtracts ``src2`` from ``src1`` and stores the the result in ``dst`` with
    the specified type.

    :param src1: source image 1
    :param src2: source image 2
    :param dst: destination image
    :param dst_type: destination image DataType
    :param saturate: whether to apply saturation
    :return: None
    """
    return _ecvl.Sub(src1, src2, dst, dst_type, saturate)


def Mul(src1, src2, dst, dst_type=DataType.none, saturate=True):
    """\
    Multiply two images.

    Muliplies ``src1`` by ``src2`` and stores the the result in ``dst`` with
    the specified type.

    :param src1: source image 1
    :param src2: source image 2
    :param dst: destination image
    :param dst_type: destination image DataType
    :param saturate: whether to apply saturation
    :return: None
    """
    return _ecvl.Mul(src1, src2, dst, dst_type, saturate)


def Div(src1, src2, dst, dst_type=DataType.none, saturate=True):
    """\
    Divide an image by another.

    Divides ``src1`` by ``src2`` and stores the the result in ``dst`` with
    the specified type.

    :param src1: source image 1
    :param src2: source image 2
    :param dst: destination image
    :param dst_type: destination image DataType
    :param saturate: whether to apply saturation
    :return: None
    """
    return _ecvl.Div(src1, src2, dst, dst_type, saturate)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .image import ColorType
from . import _core
_ecvl = _core.ecvl


__all__ = [
    "SplitType",
    "Sample",
    "Split",
    "Dataset",
]


class SplitType(_ecvl.SplitType):
    """\
    Enum class representing the supported split types.
    """
    training = _ecvl.SplitType.training
    validation = _ecvl.SplitType.validation
    test = _ecvl.SplitType.test


class Sample(_ecvl.Sample):
    r"""\
    Sample image in a dataset.

    Provides the information to describe a dataset sample.

    :var location\_: absolute path(s) of the sample (list of strings)
    :var label_path\_: absolute path of the sample's ground truth
    :var label\_: sample labels (list of integers)
    :var values\_: feature index-to-value mapping
    :var size\_: original x and y dimensions of the sample
    """
    def LoadImage(self, ctype=ColorType.BGR, is_gt=False):
        """\
        Return the dataset image for this sample.

        Opens the sample image from ``location_`` or ``label_path_``, depending
        on the ``is_gt`` parameter.

        :param ctype: ColorType of the returned image
        :param is_gt: whether to load the sample image or its ground truth
        :return: sample image
        """
        return _ecvl.Sample.LoadImage(self, ctype, is_gt)


class Split(_ecvl.Split):
    r"""\
    Provides the splits a dataset can have: training, validation, and test.

    :var training\_: training samples indices (list of integers)
    :var validation\_: validation samples indices (list of integers)
    :var test\_: test samples indices (list of integers)
    """


class Dataset(_ecvl.Dataset):
    r"""\
    DeepHealth Dataset.

    Implements the `DeepHealth Dataset Format
    <https://github.com/deephealthproject/ecvl/wiki/DeepHealth-Toolkit-Dataset-Format>`_.

    :var name\_: dataset name
    :var description\_: dataset description
    :var classes\_: classes available in the dataset (list of strings)
    :var features\_: features available in the dataset (list of strings)
    :var samples\_: list of dataset samples
    :var split\_: dataset splits
    """
    def __init__(self, filename):
        """\
        :param filename: path to the dataset file
        """
        _ecvl.Dataset.__init__(self, filename)

    def Dump(self, path):
        """\
        Dump the dataset to YAML following the DeepHealth Dataset Format.

        The YAML file is saved into the dataset's root directory. Sample paths
        are relative to the dataset's root directory.

        :param path: output file path
        :return: None
        """
        _ecvl.Dataset.Dump(self, path)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
Reads an image, increases its brightness and saves the result.
"""

import argparse
import sys

import numpy as np
import pyecvl.ecvl as ecvl


def inc_brightness(img, rate):
    a = np.array(img, copy=False)
    max_val = np.iinfo(a.dtype).max
    a[a > max_val - rate] = max_val
    a[a <= max_val - rate] += rate


def main(args):
    img = ecvl.ImRead(args.in_fn)
    inc_brightness(img, 10)
    ecvl.ImWrite(args.out_fn, img)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("in_fn", metavar="INPUT_PATH")
    parser.add_argument("out_fn", metavar="OUTPUT_PATH")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
I/O on special formats.
"""

import argparse
import os
import sys

import pyecvl.ecvl as ecvl


def nifti_io(fn):
    head, ext = os.path.splitext(os.path.basename(fn))
    print("Reading %s" % fn)
    img = ecvl.NiftiRead(fn)
    ecvl.GammaContrast(img, img, 3)
    out_fn = "%s_gamma%s" % (head, ext)
    print("Saving %s" % out_fn)
    ecvl.NiftiWrite(out_fn, img)


def dicom_io(fn):
    head, ext = os.path.splitext(os.path.basename(fn))
    print("Reading %s" % fn)
    img = ecvl.DicomRead(fn)
    ecvl.ChangeColorSpace(img, img, ecvl.ColorType.GRAY)
    thresh = ecvl.OtsuThreshold(img)
    maxval = 255
    ecvl.Threshold(img, img, thresh, maxval)
    out_fn = "%s_threshold%s" % (head, ext)
    print("Saving %s" % out_fn)
    ecvl.DicomWrite(out_fn, img)


def main(args):
    if not ecvl.ECVL_WITH_DICOM:
        print("No DICOM support - quitting")
        sys.exit(0)
    nifti_io(args.nifti_fn)
    dicom_io(args.dicom_fn)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("nifti_fn", metavar="NIFTI_INPUT_PATH")
    parser.add_argument("dicom_fn", metavar="DICOM_INPUT_PATH")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import pyecvl._core.ecvl as ecvl_core
import pyecvl.ecvl as ecvl_py

pytest.importorskip("pyecvl.augmentations")


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugmentationParam(ecvl):
    ap = ecvl.AugmentationParam()
    min_, max_ = 0.1, 1.2
    ap = ecvl.AugmentationParam(min_, max_)
    assert ap.min_ == pytest.approx(min_)
    assert ap.max_ == pytest.approx(max_)
    ecvl.AugmentationParam.SetSeed(12345)
    ap.GenerateValue()
    assert min_ <= ap.value_ <= max_


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugRotate(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugRotate([30, 50])
    a.Apply(img)
    a = ecvl.AugRotate([30, 50], [2, 3])
    a.Apply(img)
    a = ecvl.AugRotate([30, 50], [2, 3], 1.1)
    a.Apply(img)
    a = ecvl.AugRotate([30, 50], [2, 3], 1.1, ecvl.InterpolationType.nearest)
    a.Apply(img)
    a = ecvl.AugRotate([30, 50], [2, 3], 1.1, ecvl.InterpolationType.nearest,
                       ecvl.InterpolationType.linear)
    a.Apply(img)
    # fromtext
    f = ecvl.AugRotate if ecvl is ecvl_core else ecvl.AugRotate.fromtext
    a = f('angle=[30, 50] center=(2, 3) scale=1.1 interp="nearest" '
          'gt_interp="linear"')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugResizeDim(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugResizeDim([4, 3])
    a.Apply(img)
    a = ecvl.AugResizeDim([4, 3], ecvl.InterpolationType.nearest,
                          ecvl.InterpolationType.linear)
    a.Apply(img)
    # fromtext
    f = ecvl.AugResizeDim if ecvl is ecvl_core else ecvl.AugResizeDim.fromtext
    a = f('dims=(4, 3) interp="linear" gt_interp="nearest"')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugResizeScale(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugResizeScale([0.5, 0.5])
    a.Apply(img)
    a = ecvl.AugResizeScale([0.5, 0.5], ecvl.InterpolationType.nearest,
                            ecvl.InterpolationType.linear)
    a.Apply(img)
    # fromtext
    f = ecvl.AugResizeScale if ecvl is ecvl_core else \
        ecvl.AugResizeScale.fromtext
    a = f('scale=(0.5, 0.5) interp="linear" gt_interp="nearest"')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugFlip(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugFlip(0.5)
    a.Apply(img)
    # fromtext
    f = ecvl.AugFlip if ecvl is ecvl_core else ecvl.AugFlip.fromtext
    a = f('p=0.5')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugMirror(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugMirror(0.5)
    a.Apply(img)
    # fromtext
    f = ecvl.AugMirror if ecvl is ecvl_core else ecvl.AugMirror.fromtext
    a = f('p=0.5')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugGaussianBlur(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugGaussianBlur([0.2, 0.4])
    a.Apply(img)
    # fromtext
    f = ecvl.AugGaussianBlur if ecvl is ecvl_core else \
        ecvl.AugGaussianBlur.fromtext
    a = f('sigma=[0.2, 0.4]')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugAdditiveLaplaceNoise(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugAdditiveLaplaceNoise([255 * 0.05, 255 * 0.09])
    a.Apply(img)
    # fromtext
    f = ecvl.AugAdditiveLaplaceNoise if ecvl is ecvl_core else \
        ecvl.AugAdditiveLaplaceNoise.fromtext
    a = f('std_dev=[12.5, 23.1]')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugAdditivePoissonNoise(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugAdditivePoissonNoise([2.0, 3.0])
    a.Apply(img)
    # fromtext
    f = ecvl.AugAdditivePoissonNoise if ecvl is ecvl_core else \
        ecvl.AugAdditivePoissonNoise.fromtext
    a = f('lambda=[2.0, 3.0]')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugGammaContrast(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugGammaContrast([3, 4])
    a.Apply(img)
    # fromtext
    f = ecvl.AugGammaContrast if ecvl is ecvl_core else \
        ecvl.AugGammaContrast.fromtext
    a = f('gamma=[3, 4]')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugCoarseDropout(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugCoarseDropout([0.5, 0.7], [0.1, 0.2], 0.4)
    a.Apply(img)
    # fromtext
    f = ecvl.AugCoarseDropout if ecvl is ecvl_core else \
        ecvl.AugCoarseDropout.fromtext
    a = f('p=[0.5, 0.7] drop_size=[0.1, 0.2] per_channel=0.4')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugTranspose(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugTranspose(0.5)
    a.Apply(img)
    # fromtext
    f = ecvl.AugTranspose if ecvl is ecvl_core else ecvl.AugTranspose.fromtext
    a = f('p=0.5')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugBrightness(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugBrightness([30, 60])
    a.Apply(img)
    # fromtext
    f = ecvl.AugBrightness if ecvl is ecvl_core else \
        ecvl.AugBrightness.fromtext
    a = f('beta=[30, 60]')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugGridDistortion(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugGridDistortion([5, 10], [-0.2, 0.2])
    a.Apply(img)
    a = ecvl.AugGridDistortion(
        [5, 10], [-0.2, 0.2], ecvl.InterpolationType.nearest
    )
    a.Apply(img)
    a = ecvl.AugGridDistortion(
        [5, 10], [-0.2, 0.2], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_REFLECT_101
    )
    a.Apply(img)
    a = ecvl.AugGridDistortion(
        [5, 10], [-0.2, 0.2], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_REFLECT_101, 0
    )
    a.Apply(img)
    # fromtext
    f = ecvl.AugGridDistortion if ecvl is ecvl_core else \
        ecvl.AugGridDistortion.fromtext
    a = f('num_steps=[5,10] distort_limit=[-0.2,0.2] interp=\"linear\" '
          'border_type=\"reflect_101\" border_value=0')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugOpticalDistortion(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugOpticalDistortion([-0.2, 0.2], [-0.4, 0.4])
    a.Apply(img)
    a = ecvl.AugOpticalDistortion(
        [-0.2, 0.2], [-0.4, 0.4], ecvl.InterpolationType.nearest
    )
    a.Apply(img)
    a = ecvl.AugOpticalDistortion(
        [-0.2, 0.2], [-0.4, 0.4], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_REFLECT_101
    )
    a.Apply(img)
    a = ecvl.AugOpticalDistortion(
        [-0.2, 0.2], [-0.4, 0.4], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_REFLECT_101, 0
    )
    a.Apply(img)
    # fromtext
    f = ecvl.AugOpticalDistortion if ecvl is ecvl_core else \
        ecvl.AugOpticalDistortion.fromtext
    a = f('distort_limit=[-0.2,0.2] shift_limit=[-0.4,0.4] interp=\"linear\" '
          'border_type=\"reflect_101\" border_value=0')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugSalt(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugSalt([0.1, 0.3], 0.5)
    a.Apply(img)
    # fromtext
    f = ecvl.AugSalt if ecvl is ecvl_core else ecvl.AugSalt.fromtext
    a = f('p=[0.1,0.3] per_channel=0.5')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugPepper(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugPepper([0.1, 0.3], 0.5)
    a.Apply(img)
    # fromtext
    f = ecvl.AugPepper if ecvl is ecvl_core else ecvl.AugPepper.fromtext
    a = f('p=[0.1,0.3] per_channel=0.5')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugSaltAndPepper(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugSaltAndPepper([0.1, 0.3], 0.5)
    a.Apply(img)
    # fromtext
    f = ecvl.AugSaltAndPepper if ecvl is ecvl_core else \
        ecvl.AugSaltAndPepper.fromtext
    a = f('p=[0.1,0.3] per_channel=0.5')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugElasticTransform(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugElasticTransform([34, 60], [4, 6])
    a.Apply(img)
    a = ecvl.AugElasticTransform(
        [34, 60], [4, 6], ecvl.InterpolationType.nearest
    )
    a.Apply(img)
    a = ecvl.AugElasticTransform(
        [34, 60], [4, 6], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_REFLECT_101
    )
    a.Apply(img)
    a = ecvl.AugElasticTransform(
        [34, 60], [4, 6], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_REFLECT_101, 0
    )
    a.Apply(img)
    # fromtext
    f = ecvl.AugElasticTransform if ecvl is ecvl_core else \
        ecvl.AugElasticTransform.fromtext
    a = f('alpha=[34,60] sigma=[4,6] interp=\"linear\" '
          'border_type=\"reflect_101\" border_value=0')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugNormalize(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugNormalize(20, 5.5)
    a.Apply(img)
    # fromtext
    f = ecvl.AugNormalize if ecvl is ecvl_core else \
        ecvl.AugNormalize.fromtext
    a = f('mean=20 std=5.5')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugNormalize_separate(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugNormalize([20, 19, 21], [5, 5.5, 6])
    a.Apply(img)
    # fromtext
    f = ecvl.AugNormalize if ecvl is ecvl_core else \
        ecvl.AugNormalize.fromtext
    a = f('mean=(20,19,21) std=(5,5.5,6)')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugCenterCrop(ecvl):
    img = ecvl.Image([8, 6, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugCenterCrop([4, 3])
    a.Apply(img)
    # fromtext
    f = ecvl.AugCenterCrop if ecvl is ecvl_core else \
        ecvl.AugCenterCrop.fromtext
    a = f('size=(4,3)')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugToFloat32(ecvl):
    img = ecvl.Image([8, 6, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugToFloat32()
    a.Apply(img)
    assert img.elemtype_ == ecvl.DataType.float32
    img = ecvl.Image([8, 6, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugToFloat32(2.)
    a.Apply(img)
    assert img.elemtype_ == ecvl.DataType.float32
    img = ecvl.Image([8, 6, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugToFloat32(2., 3.)
    a.Apply(img)
    assert img.elemtype_ == ecvl.DataType.float32
    # fromtext
    f = ecvl.AugToFloat32 if ecvl is ecvl_core else \
        ecvl.AugToFloat32.fromtext
    img = ecvl.Image([8, 6, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = f('divisor=2. divisor_gt=3.')
    a.Apply(img)
    assert img.elemtype_ == ecvl.DataType.float32


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugDivBy255(ecvl):
    img = ecvl.Image([8, 6, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugDivBy255()
    a.Apply(img)
    # fromtext
    f = ecvl.AugDivBy255 if ecvl is ecvl_core else \
        ecvl.AugDivBy255.fromtext
    a = f('')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugScaleTo(ecvl):
    img = ecvl.Image([8, 6, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    a = ecvl.AugScaleTo(1, 254)
    a.Apply(img)
    # fromtext
    f = ecvl.AugScaleTo if ecvl is ecvl_core else \
        ecvl.AugScaleTo.fromtext
    a = f('new_min=1 new_max=255')
    a.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AugmentationFactory(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    # one arg
    a = ecvl.AugmentationFactory.create('AugFlip p=0.5')
    a.Apply(img)
    # two args
    a = ecvl.AugmentationFactory.create('AugFlip', 'p=0.5')
    a.Apply(img)
    # container
    txt = ('SequentialAugmentationContainer\n'
           'AugRotate angle=[-5,5] center=(0,0) scale=0.5 interp="linear"\n'
           'AugAdditiveLaplaceNoise std_dev=[0,0.51]\n'
           'AugCoarseDropout p=[0,0.55] drop_size=[0.02,0.1] per_channel=0\n'
           'AugAdditivePoissonNoise lambda=[0,40]\n'
           'AugResizeDim dims=(30,30) interp="linear"\n'
           'end')
    c = ecvl.AugmentationFactory.create(txt)
    c.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_SequentialAugmentationContainer(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    # from list
    c = ecvl.SequentialAugmentationContainer([
        ecvl.AugRotate([-5, 5]),
        ecvl.AugMirror(.5),
    ])
    c.Apply(img)
    # fromtext
    txt = ('AugFlip p=0.2\n'
           'AugMirror p=0.2\n'
           'end')
    f = ecvl.SequentialAugmentationContainer if ecvl is ecvl_core else \
        ecvl.SequentialAugmentationContainer.fromtext
    c = f(txt)
    c.Apply(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_OneOfAugmentationContainer(ecvl):
    img = ecvl.Image([5, 4, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    # from list
    c = ecvl.OneOfAugmentationContainer(0.7, [
        ecvl.AugRotate([-5, 5]),
        ecvl.AugMirror(.5),
    ])
    c.Apply(img)
    # fromtext
    txt = ('p=0.7\n'
           'AugFlip p=0.2\n'
           'AugMirror p=0.2\n'
           'end')
    f = ecvl.OneOfAugmentationContainer if ecvl is ecvl_core else \
        ecvl.OneOfAugmentationContainer.fromtext
    c = f(txt)
    c.Apply(img)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import pytest

import pyecvl._core.ecvl as ecvl_core
import pyecvl.ecvl as ecvl_py


def _empty_img(ecvl):
    if ecvl is ecvl_core:
        return ecvl.Image()
    return ecvl.Image.empty()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ResizeDim(ecvl):
    dims = [20, 40, 3]
    newdims = [10, 20]  # no color channel
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.ResizeDim(img, tmp, newdims)
    assert tmp.dims_[:2] == newdims
    ecvl.ResizeDim(img, tmp, newdims, ecvl.InterpolationType.nearest)
    assert tmp.dims_[:2] == newdims


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ResizeScale(ecvl):
    dims = [20, 40, 3]
    scales = [0.5, 0.5]  # no color channel
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.ResizeScale(img, tmp, scales)
    assert tmp.dims_[:2] == [10, 20]
    ecvl.ResizeScale(img, tmp, scales, ecvl.InterpolationType.cubic)
    assert tmp.dims_[:2] == [10, 20]


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Flip2D(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.Flip2D(img, tmp)
    assert tmp.dims_ == img.dims_


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Mirror2D(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.Mirror2D(img, tmp)
    assert tmp.dims_ == img.dims_


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Rotate2D(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    angle, center, scale, interp = 9, [5, 5], 1.5, ecvl.InterpolationType.area
    ecvl.Rotate2D(img, tmp, angle)
    ecvl.Rotate2D(img, tmp, angle, center)
    ecvl.Rotate2D(img, tmp, angle, center, scale)
    ecvl.Rotate2D(img, tmp, angle, center, scale, interp)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_RotateFullImage2D(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    angle, scale, interp = 9, 1.5, ecvl.InterpolationType.lanczos4
    ecvl.RotateFullImage2D(img, tmp, angle)
    ecvl.RotateFullImage2D(img, tmp, angle, scale)
    ecvl.RotateFullImage2D(img, tmp, angle, scale, interp)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ChangeColorSpace(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    new_color = ecvl.ColorType.GRAY
    ecvl.ChangeColorSpace(img, tmp, new_color)
    assert tmp.colortype_ == new_color
    assert tmp.dims_[-1] == 1


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Threshold(ecvl):
    dims = [20, 40, 1]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.GRAY)
    thr = ecvl.OtsuThreshold(img)
    tmp = _empty_img(ecvl)
    ttype = ecvl.ThresholdingType.BINARY_INV
    ecvl.Threshold(img, tmp, thr, 255)
    ecvl.Threshold(img, tmp, thr, 255, ttype)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_MultiThreshold(ecvl):
    dims = [20, 40, 1]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.GRAY)
    thresholds = ecvl.OtsuMultiThreshold(img)
    tmp = _empty_img(ecvl)
    ecvl.MultiThreshold(img, tmp, thresholds)
    ecvl.MultiThreshold(img, tmp, thresholds, 1)
    ecvl.MultiThreshold(img, tmp, thresholds, 1, 254)
    thresholds = ecvl.OtsuMultiThreshold(img, 3)
    tmp = _empty_img(ecvl)
    ecvl.MultiThreshold(img, tmp, thresholds)
    ecvl.MultiThreshold(img, tmp, thresholds, 1)
    ecvl.MultiThreshold(img, tmp, thresholds, 1, 254)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Filter2D(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    # kernel must be float64, "xyc" and with one color channel
    kernel = ecvl.Image(
        [3, 3, 1], ecvl.DataType.float64, "xyc", ecvl.ColorType.GRAY
    )
    a = np.array(kernel, copy=False)
    a.fill(0.11)
    dtype = ecvl.DataType.uint16
    ecvl.Filter2D(img, tmp, kernel)
    ecvl.Filter2D(img, tmp, kernel, dtype)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_SeparableFilter2D(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    kerX, kerY, dtype = [1, 2, 1], [1, 0, -1], ecvl.DataType.uint16
    ecvl.SeparableFilter2D(img, tmp, kerX, kerY)
    ecvl.SeparableFilter2D(img, tmp, kerX, kerY, dtype)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_GaussianBlur(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    sigmaY = 0.2
    ecvl.GaussianBlur(img, tmp, 5, 5, 0.1)
    ecvl.GaussianBlur(img, tmp, 5, 5, 0.1, sigmaY)
    # alt overload in the ext module, called "GaussianBlur2" in the wrapper
    GaussianBlur2 = getattr(ecvl, "GaussianBlur2", ecvl.GaussianBlur)
    GaussianBlur2(img, tmp, 0.2)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AdditiveLaplaceNoise(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    stddev = 255 * 0.05
    ecvl.AdditiveLaplaceNoise(img, tmp, stddev)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_AdditivePoissonNoise(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    lambda_ = 2.0
    ecvl.AdditivePoissonNoise(img, tmp, lambda_)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_GammaContrast(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    gamma = 3
    ecvl.GammaContrast(img, tmp, gamma)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_CoarseDropout(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    prob, drop_size, per_channel = 0.5, 0.1, True
    ecvl.CoarseDropout(img, tmp, prob, drop_size, per_channel)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_IntegralImage(ecvl):
    dims = [20, 40, 1]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.GRAY)
    tmp = _empty_img(ecvl)
    dst_type = ecvl.DataType.float64
    ecvl.IntegralImage(img, tmp)
    ecvl.IntegralImage(img, tmp, dst_type)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_NonMaximaSuppression(ecvl):
    dims = [20, 40, 1]
    img = ecvl.Image(dims, ecvl.DataType.int32, "xyc", ecvl.ColorType.GRAY)
    tmp = _empty_img(ecvl)
    ecvl.NonMaximaSuppression(img, tmp)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_GetMaxN(ecvl):
    a = np.asfortranarray(np.zeros(12, dtype=np.int32).reshape(3, 4, 1))
    a[0, 1] = 3
    a[1, 2] = 4
    if ecvl is ecvl_core:
        img = ecvl.Image(a, "xyc", ecvl.ColorType.GRAY)
    else:
        img = ecvl.Image.fromarray(a, "xyc", ecvl.ColorType.GRAY)
    assert sorted(ecvl.GetMaxN(img, 2)) == [[0, 1], [1, 2]]


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ConnectedComponentsLabeling(ecvl):
    dims = [20, 40, 1]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.GRAY)
    tmp = _empty_img(ecvl)
    ecvl.ConnectedComponentsLabeling(img, tmp)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_FindContours(ecvl):
    dims = [20, 40, 1]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.GRAY)
    ecvl.FindContours(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_HConcat(ecvl):
    img1 = ecvl.Image(
        [20, 40, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    img2 = ecvl.Image(
        [40, 40, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    tmp = _empty_img(ecvl)
    ecvl.HConcat([img1, img2], tmp)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_VConcat(ecvl):
    img1 = ecvl.Image(
        [20, 40, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    img2 = ecvl.Image(
        [20, 20, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    tmp = _empty_img(ecvl)
    ecvl.VConcat([img1, img2], tmp)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Stack(ecvl):
    img1 = ecvl.Image(
        [20, 40, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    img2 = ecvl.Image(
        [20, 40, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    tmp = _empty_img(ecvl)
    ecvl.Stack([img1, img2], tmp)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Morphology(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    kernel = ecvl.Image(
        [5, 5, 1], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    tmp = _empty_img(ecvl)
    ecvl.Morphology(img, tmp, ecvl.MorphType.MORPH_BLACKHAT, kernel)
    ecvl.Morphology(img, tmp, ecvl.MorphType.MORPH_BLACKHAT, kernel, [3, 3])
    ecvl.Morphology(
        img, tmp, ecvl.MorphType.MORPH_BLACKHAT, kernel, [3, 3], 2
    )
    ecvl.Morphology(
        img, tmp, ecvl.MorphType.MORPH_BLACKHAT, kernel, [3, 3], 2,
        ecvl.BorderType.BORDER_REFLECT
    )
    ecvl.Morphology(
        img, tmp, ecvl.MorphType.MORPH_BLACKHAT, kernel, [3, 3], 2,
        ecvl.BorderType.BORDER_REFLECT, 1
    )


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Inpaint(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    mask = ecvl.Image(
        [20, 40, 1], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    tmp = _empty_img(ecvl)
    ecvl.Inpaint(img, tmp, mask, 5.0)
    ecvl.Inpaint(img, tmp, mask, 5.0, ecvl.InpaintType.INPAINT_NS)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_MeanStdDev(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    mean, stddev = ecvl.MeanStdDev(img)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Transpose(ecvl):
    img = ecvl.Image(
        [20, 40, 3], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    tmp = _empty_img(ecvl)
    ecvl.Transpose(img, tmp)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_GridDistortion(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.GridDistortion(img, tmp)
    ecvl.GridDistortion(img, tmp, 6)
    ecvl.GridDistortion(img, tmp, 6, [-0.5, 0.5])
    ecvl.GridDistortion(
        img, tmp, 6, [-0.5, 0.5], ecvl.InterpolationType.nearest
    )
    ecvl.GridDistortion(
        img, tmp, 6, [-0.5, 0.5], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_CONSTANT
    )
    ecvl.GridDistortion(
        img, tmp, 6, [-0.5, 0.5], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_CONSTANT, 1
    )
    ecvl.GridDistortion(
        img, tmp, 6, [-0.5, 0.5], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_CONSTANT, 1, 12345
    )


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ElasticTransform(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.ElasticTransform(img, tmp)
    ecvl.ElasticTransform(img, tmp, 35)
    ecvl.ElasticTransform(img, tmp, 35, 5)
    ecvl.ElasticTransform(img, tmp, 35, 5, ecvl.InterpolationType.nearest)
    ecvl.ElasticTransform(
        img, tmp, 35, 5, ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_CONSTANT
    )
    ecvl.ElasticTransform(
        img, tmp, 35, 5, ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_CONSTANT, 1
    )
    ecvl.ElasticTransform(
        img, tmp, 35, 5, ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_CONSTANT, 1, 12345
    )


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_OpticalDistortion(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.OpticalDistortion(img, tmp)
    ecvl.OpticalDistortion(img, tmp)
    ecvl.OpticalDistortion(img, tmp, [-0.5, 0.5])
    ecvl.OpticalDistortion(img, tmp, [-0.5, 0.5], [-0.2, 0.2])
    ecvl.OpticalDistortion(
        img, tmp, [-0.5, 0.5], [-0.2, 0.2], ecvl.InterpolationType.nearest
    )
    ecvl.OpticalDistortion(
        img, tmp, [-0.5, 0.5], [-0.2, 0.2], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_CONSTANT
    )
    ecvl.OpticalDistortion(
        img, tmp, [-0.5, 0.5], [-0.2, 0.2], ecvl.InterpolationType.nearest,
        ecvl.BorderType.BORDER_CONSTANT, 12345
    )


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Salt(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.Salt(img, tmp, 0.4)
    ecvl.Salt(img, tmp, 0.4, True)
    ecvl.Salt(img, tmp, 0.4, True, 12345)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Pepper(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.Pepper(img, tmp, 0.4)
    ecvl.Pepper(img, tmp, 0.4, True)
    ecvl.Pepper(img, tmp, 0.4, True, 12345)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_SaltAndPepper(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.SaltAndPepper(img, tmp, 0.4)
    ecvl.SaltAndPepper(img, tmp, 0.4, True)
    ecvl.SaltAndPepper(img, tmp, 0.4, True, 12345)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_SliceTimingCorrection(ecvl):
    dims = [20, 40, 10, 30]
    spacings = [1, 2, 3, 4]
    img = ecvl.Image(
        dims, ecvl.DataType.float32, "xyzt", ecvl.ColorType.none, spacings
    )
    tmp = _empty_img(ecvl)
    ecvl.SliceTimingCorrection(img, tmp)
    ecvl.SliceTimingCorrection(img, tmp, True)
    ecvl.SliceTimingCorrection(img, tmp, True, True)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_DrawEllipse(ecvl):
    dims = [20, 20, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    ecvl.DrawEllipse(img, [10, 10], [7, 5], 15, [100])
    ecvl.DrawEllipse(img, [10, 10], [7, 5], 15, [50, 50, 50])
    ecvl.DrawEllipse(img, [10, 10], [7, 5], 15, [50, 50, 50], 2)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Moments(ecvl):
    dims = [20, 20, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.GRAY)
    moments = _empty_img(ecvl)
    ecvl.Moments(img, moments)
    assert moments.dims_ == [4, 4]
    assert moments.elemtype_ == ecvl.DataType.float64
    ecvl.Moments(img, moments, 2)
    assert moments.dims_ == [3, 3]
    assert moments.elemtype_ == ecvl.DataType.float64
    datatype = ecvl.DataType.float32
    ecvl.Moments(img, moments, 2, datatype)
    assert moments.dims_ == [3, 3]
    assert moments.elemtype_ == datatype


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_CentralMoments(ecvl):
    dims = [20, 20, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.GRAY)
    moments = _empty_img(ecvl)
    ecvl.CentralMoments(img, moments, [10., 10.])
    assert moments.dims_ == [4, 4]
    assert moments.elemtype_ == ecvl.DataType.float64
    ecvl.CentralMoments(img, moments, [10., 10.], 2)
    assert moments.dims_ == [3, 3]
    assert moments.elemtype_ == ecvl.DataType.float64
    datatype = ecvl.DataType.float32
    ecvl.CentralMoments(img, moments, [10., 10.], 2, datatype)
    assert moments.dims_ == [3, 3]
    assert moments.elemtype_ == datatype


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_DropColorChannel(ecvl):
    dims = [20, 20, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.GRAY)
    ecvl.DropColorChannel(img)
    assert img.dims_ == [20, 20]
    assert img.channels_ == "xy"
    assert img.colortype_ == ecvl.ColorType.none


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Normalize(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.Normalize(img, tmp, 20, 5.5)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Normalize_separate(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.Normalize(img, tmp, [20, 19, 21], [5, 5.5, 6])


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_CenterCrop(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.CenterCrop(img, tmp, [10, 20])


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ScaleTo(ecvl):
    dims = [20, 40, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    tmp = _empty_img(ecvl)
    ecvl.ScaleTo(img, tmp, 1, 254)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import _core
_ecvl = _core.ecvl


if not _ecvl.ECVL_WITH_DICOM:
    raise ImportError("extension module not compiled with DCMTK support")


__all__ = [
    "DicomRead",
    "DicomWrite",
]


def DicomRead(filename):
    """\
    Loads a dicom image from a file.

    :param filename: image file name
    :return: an Image object
    """
    return _ecvl.DicomRead(filename)


def DicomWrite(filename, src):
    """\
    Save an image to a file in the DICOM format.

    :param filename: image file name
    :param src: image to be saved
    :return: None
    """
    return _ecvl.DicomWrite(filename, src)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import pytest

import pyecvl._core.ecvl as ecvl_core
import pyecvl.ecvl as ecvl_py


def _empty_img(ecvl):
    if ecvl is ecvl_core:
        return ecvl.Image()
    return ecvl.Image.empty()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Neg(ecvl):
    dims = [2, 3, 1]
    z = ecvl.Image(dims, ecvl.DataType.int8, "xyc", ecvl.ColorType.GRAY)
    d = np.array(z, copy=False)
    d.fill(10)
    tmp = _empty_img(ecvl)
    # basic call
    ecvl.Neg(z, tmp)
    tmp_arr = np.array(tmp, copy=False)
    assert (tmp_arr == -10).all()
    # data type
    d.fill(20)
    ecvl.Neg(z, tmp, ecvl.DataType.int16)
    tmp_arr = np.array(tmp, copy=False)
    assert (tmp_arr == -20).all()
    assert (tmp_arr.dtype == np.int16)
    # saturate
    d.fill(-128)
    ecvl.Neg(z, tmp, ecvl.DataType.int8, True)
    tmp_arr = np.array(tmp, copy=False)
    assert (tmp_arr == 127).all()
    ecvl.Neg(z, tmp, ecvl.DataType.int8, False)
    tmp_arr = np.array(tmp, copy=False)
    assert (tmp_arr == -128).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_Add(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    y = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    b = np.array(y, copy=False)
    a.fill(10)
    b.fill(3)
    z = _empty_img(ecvl)
    # basic call
    ecvl.Add(x, y, z)
    c = np.array(z, copy=False)
    assert (c == 13).all()
    # data type
    a.fill(20)
    ecvl.Add(x, y, z, ecvl.DataType.int16)
    c = np.array(z, copy=False)
    assert (c == 23).all()
    assert (c.dtype == np.int16)
    # saturate
    a.fill(10)
    b.fill(118)
    ecvl.Add(x, y, z, ecvl.DataType.int8, True)
    c = np.array(z, copy=False)
    assert (c == 127).all()
    ecvl.Add(x, y, z, ecvl.DataType.int8, False)
    c = np.array(z, copy=False)
    assert (c == -128).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_sub(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    y = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    b = np.array(y, copy=False)
    a.fill(-10)
    b.fill(3)
    z = _empty_img(ecvl)
    # basic call
    ecvl.Sub(x, y, z)
    c = np.array(z, copy=False)
    assert (c == -13).all()
    # data type
    a.fill(-20)
    ecvl.Sub(x, y, z, ecvl.DataType.int16)
    c = np.array(z, copy=False)
    assert (c == -23).all()
    assert (c.dtype == np.int16)
    # saturate
    a.fill(-10)
    b.fill(119)
    ecvl.Sub(x, y, z, ecvl.DataType.int8, True)
    c = np.array(z, copy=False)
    assert (c == -128).all()
    ecvl.Sub(x, y, z, ecvl.DataType.int8, False)
    c = np.array(z, copy=False)
    assert (c == 127).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_mul(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    y = ecvl.Image([2, 4, 3], ecvl.DataType.int8, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    b = np.array(y, copy=False)
    a.fill(16)
    b.fill(3)
    z = _empty_img(ecvl)
    # basic call
    ecvl.Mul(x, y, z)
    c = np.array(z, copy=False)
    assert (c == 48).all()
    # data type
    a.fill(20)
    ecvl.Mul(x, y, z, ecvl.DataType.int16)
    c = np.array(z, copy=False)
    assert (c == 60).all()
    assert (c.dtype == np.int16)
    # saturate
    a.fill(16)
    b.fill(8)
    ecvl.Mul(x, y, z, ecvl.DataType.int8, True)
    c = np.array(z, copy=False)
    assert (c == 127).all()
    ecvl.Mul(x, y, z, ecvl.DataType.int8, False)
    c = np.array(z, copy=False)
    assert (c == -128).all()


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_div(ecvl):
    x = ecvl.Image([2, 4, 3], ecvl.DataType.int16, "xyc", ecvl.ColorType.RGB)
    y = ecvl.Image([2, 4, 3], ecvl.DataType.int16, "xyc", ecvl.ColorType.RGB)
    a = np.array(x, copy=False)
    b = np.array(y, copy=False)
    a.fill(11)
    b.fill(2)
    z = _empty_img(ecvl)
    # basic call
    ecvl.Div(x, y, z)
    c = np.array(z, copy=False)
    assert (c == 5).all()
    # data type
    a.fill(20)
    ecvl.Div(x, y, z, ecvl.DataType.int8)
    c = np.array(z, copy=False)
    assert (c == 10).all()
    assert (c.dtype == np.int8)
    # saturate
    a.fill(1280)
    b.fill(10)
    ecvl.Div(x, y, z, ecvl.DataType.int8, True)
    c = np.array(z, copy=False)
    # assert (c == 127).all()  # bug in ecvl::Div?
    ecvl.Div(x, y, z, ecvl.DataType.int8, False)
    c = np.array(z, copy=False)
    # assert (c == -128).all()  # bug in ecvl::Div?



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
import numpy as np

import pyecvl._core.ecvl as ecvl_core
import pyecvl.ecvl as ecvl_py
tensor = pytest.importorskip("pyeddl.tensor")


def _img_fromarray(ecvl, array, channels, colortype, spacings=None):
    if spacings is None:
        spacings = []
    if ecvl is ecvl_core:
        return ecvl.Image(array, channels, colortype, spacings)
    return ecvl.Image.fromarray(array, channels, colortype, spacings)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ImageToTensor(ecvl):
    dims = [20, 30, 3]
    img = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    t = ecvl.ImageToTensor(img)
    t.max()
    assert t.shape == [dims[2], dims[1], dims[0]]


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_ImageToTensor_offset(ecvl):
    a = np.arange(24).reshape((3, 4, 2)).astype(np.uint8)
    b = np.arange(24, 48).reshape((3, 4, 2)).astype(np.uint8)
    a_img = _img_fromarray(ecvl, a, "xyc", ecvl.ColorType.BGR)
    b_img = _img_fromarray(ecvl, b, "xyc", ecvl.ColorType.BGR)
    t = ecvl.ImageToTensor(a_img)
    u = ecvl.ImageToTensor(b_img)
    assert t.shape == [2, 4, 3]
    t_arr = t.getdata()
    assert u.shape == [2, 4, 3]
    u_arr = u.getdata()
    v = tensor.Tensor([4, 4, 3])
    ecvl.ImageToTensor(a_img, v, 0)
    ecvl.ImageToTensor(b_img, v, 1)
    v_arr = v.getdata()
    assert np.array_equal(v_arr[:2], t_arr)
    assert np.array_equal(v_arr[2:], u_arr)


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_TensorToImage(ecvl):
    # 3D
    shape = [3, 30, 20]
    t = tensor.Tensor(shape)
    img = ecvl.TensorToImage(t)
    assert img.dims_ == [shape[2], shape[1], shape[0]]
    # 4D
    shape = [5, 3, 30, 20]
    t = tensor.Tensor(shape)
    img = ecvl.TensorToImage(t)
    assert img.dims_ == [shape[3], shape[2], shape[0] * shape[1]]


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_TensorToView(ecvl):
    # 3D
    shape = [3, 30, 20]
    t = tensor.Tensor(shape)
    view = ecvl.TensorToView(t)
    assert view.dims_ == [shape[2], shape[1], shape[0]]
    # 4D
    shape = [5, 3, 30, 20]
    t = tensor.Tensor(shape)
    view = ecvl.TensorToView(t)
    assert view.dims_ == [shape[3], shape[2], shape[0] * shape[1]]



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import _core
_ecvl = _core.ecvl


__all__ = [
    "ColorType",
    "Device",
    "Image",
    "View_int8",
    "View_int16",
    "View_float32",
    "CopyImage",
    "ShallowCopyImage",
    "RearrangeChannels",
    "ConvertTo",
]


class ColorType(_ecvl.ColorType):
    """\
    Enum class representing the supported color spaces.
    """
    none = _ecvl.ColorType.none
    GRAY = _ecvl.ColorType.GRAY
    RGB = _ecvl.ColorType.RGB
    RGBA = _ecvl.ColorType.RGBA
    BGR = _ecvl.ColorType.BGR
    HSV = _ecvl.ColorType.HSV
    YCbCr = _ecvl.ColorType.YCbCr


class Device(_ecvl.Device):
    """\
    Enum class representing the supported devices.
    """
    NONE = _ecvl.Device.NONE
    CPU = _ecvl.Device.CPU
    GPU = _ecvl.Device.GPU
    FPGA = _ecvl.Device.FPGA


class Image(_ecvl.Image):
    r"""\
    Image class.

    :var elemtype\_: image pixel type, a DataType

    :var elemsize\_: image pixel size in bytes

    :dims\_: image dimensions in pixels/voxels (list of integers)

    :strides\_: the number of bytes the internal data pointer has to skip to
      get to the next pixel/voxel in the corresponding dimension

    :var channels\_: a string describing the image planes layout. Each
      character provides information on the corresponding channel. The
      possible values are: 'x': horizontal spatial dimension; 'y': vertical
      spatial dimension; 'z': depth spatial dimension; 'c': color dimension;
      't': temporal dimension; 'o': any other dimension. For example, "xyc"
      describes a 2-dimensional image structured in color planes. This could
      be for example a ColorType.GRAY Image with dims\_[2] = 1 or a
      ColorType.RGB Image with dims_[2] = 3. The ColorType constrains the
      value of the dimension corresponding to the color channel. Another
      example is "cxy" with dims\_[0] = 3 and ColorType.BGR. In this case the
      color dimension is the one which changes faster, as it happens in
      libraries such as OpenCV.

    :var colortype\_: image ColorType. If different from ColorType.none, the
      channels\_ string must contain a 'c' and the corresponding dimension must
      have the appropriate value.

    :var spacings\_: space in mm between consecutive pixels/voxels on each
      axis (list of floats).

    :var datasize\_: size of image data in bytes.

    :var contiguous\_: whether the image is stored contiguously in memory

    :var dev\_: image Device
    """

    @staticmethod
    def empty():
        """\
        Create an empty image.

        :return: an empty image
        """
        return _ecvl.Image()

    @staticmethod
    def fromarray(array, channels, colortype, spacings=None):
        """\
        Create an image from a NumPy array.

        :param array: a NumPy array
        :param channels: channels string
        :param colortype: a ColorType
        :param spacings: spacings between pixels
        :return: an image containing the same data as the input array
        """
        if spacings is None:
            spacings = []
        return _ecvl.Image(array, channels, colortype, spacings)

    def __init__(self, dims, elemtype, channels, colortype, spacings=None,
                 dev=Device.CPU):
        """\
        :param dims: image dimensions
        :param elemtype: pixel type, a DataType
        :param channels: channels string
        :param colortype: a ColorType
        :param spacings: spacings between pixels
        :param dev: image Device
        """
        if spacings is None:
            spacings = []
        _ecvl.Image.__init__(
            self, dims, elemtype, channels, colortype, spacings, dev
        )

    def copy(self):
        """\
        Create a deep copy of this image.

        :return: a copy of this image
        """
        return _ecvl.Image(self)

    def IsEmpty(self):
        """\
        Whether the image contains data or not.

        :return: True if the image contains data, False otherwise
        """
        return _ecvl.Image.IsEmpty(self)

    def IsOwner(self):
        """\
        Whether the image owns the data or not.

        :return: True if the image owns the data, False otherwise
        """
        return _ecvl.Image.IsOwner(self)

    def Width(self):
        """\
        Get the image width.

        :return: image width
        """
        return _ecvl.Image.Width(self)

    def Height(self):
        """\
        Get the image height.

        :return: image height
        """
        return _ecvl.Image.Height(self)

    def Channels(self):
        """\
        Get the number of channels.

        :return: number of channels
        """
        return _ecvl.Image.Channels(self)

    def Add(self, other, saturate=True):
        """\
        Add data from another image to this image's data.

        :param other: other image
        :param saturate: in case of overflow, set values to limit for data type
        :return: None
        """
        return _ecvl.Image.Add(self, other, saturate)

    def Sub(self, other, saturate=True):
        """\
        Subtract data from another image from this image's data.

        :param other: other image
        :param saturate: in case of overflow, set values to limit for data type
        :return: None
        """
        return _ecvl.Image.Sub(self, other, saturate)

    def Mul(self, other, saturate=True):
        """\
        Multiply data from this image by another image's data.

        :param other: other image
        :param saturate: in case of overflow, set values to limit for data type
        :return: None
        """
        return _ecvl.Image.Mul(self, other, saturate)

    def Div(self, other, saturate=True):
        """\
        Divide data from this image by another image's data.

        :param other: other image
        :param saturate: in case of overflow, set values to limit for data type
        :return: None
        """
        return _ecvl.Image.Div(self, other, saturate)

    def To(self, dev):
        """\
        Change the image device.

        :param dev: new Device
        :return: None
        """
        return _ecvl.Image.To(self, dev)

    def ConvertTo(self, dtype, saturate=True):
        """\
        Convert Image to another DataType.

        :param dtype: target DataType
        :param saturate: whether to apply saturation or not
        """
        return _ecvl.Image.ConvertTo(self, dtype, saturate)


class View_int8(_ecvl.View_int8):
    pass


class View_int16(_ecvl.View_int16):
    pass


class View_float32(_ecvl.View_float32):
    pass


def CopyImage(src, dst, new_type=None):
    """\
    Copy data from the source image ``src`` to the destination image ``dst``.

    ``src`` and ``dst`` cannot be the same image. ``src`` cannot be an image
    with DataType.none. The optional ``new_type`` parameter can be used to
    change the DataType of the ``dst`` image.

    When the DataType is not specified:

    * if ``dst`` is empty, ``src`` will be directly copied into ``dst``

    * if ``src`` and ``dst`` have different size in memory or different
      channels and ``dst`` is the data owner, the procedure will overwrite
      ``dst`` creating a new image (channels and dimensions will be the same
      as ``src``, DataType will be the same as ``dst`` if they are not none or
      the same as ``src`` otherwise)

    * if ``src`` and ``dst`` have different size in memory or different
      channels and ``dst`` is not the data owner, the procedure will throw an
      exception

    * if ``src`` and ``dst`` have different color types and ``dst`` is the
      data owner, the procedure produces a ``dst`` image with the same color
      type as ``src``

    * if ``src`` and ``dst`` have different color types and ``dst`` is not the
      data owner, the procedure will throw an exception

    When the DataType is specified the function has the same behavior, but
    ``dst`` will have the specified DataType.

    :param src: source image
    :param dst: destination image
    :param new_type: new DataType for the destination image
    :return: None
    """
    if new_type is None:
        return _ecvl.CopyImage(src, dst)
    return _ecvl.CopyImage(src, dst, new_type)


def ShallowCopyImage(src, dst):
    """\
    Shallow copy of ``src`` to ``dst`` (``dst`` will point to the same data).

    ``src`` and ``dst`` cannot be the same image. Even though ``dst`` will
    point to the same data as ``src``, the latter will be the data owner.

    :param src: source image
    :param dst: destination image
    :return: None
    """
    return _ecvl.ShallowCopyImage(src, dst)


def RearrangeChannels(src, dst, channels, new_type=None):
    """\
    Change image dimensions order.

    Changes the order of the ``src`` image dimensions, saving the result into
    the ``dst`` image. The new order can be specified as a string through the
    ``channels`` parameter. ``src`` and ``dst`` can be the same image.

    :param src: source image
    :param dst: destination image
    :param channels: new order for the image channels, as a string
    :param new_type: new DataType for the destination image. If None, the
      destination image will preserve its type if it is not empty, otherwise
      it will have the same type as the source image
    """
    if new_type is None:
        return _ecvl.RearrangeChannels(src, dst, channels)
    return _ecvl.RearrangeChannels(src, dst, channels, new_type)


def ConvertTo(src, dst, dtype, saturate=True):
    """\
    Convert Image to another DataType.

    :param src: source image
    :param dst: destination image
    :param dtype: target DataType
    :param saturate: whether to apply saturation or not
    """
    return _ecvl.ConvertTo(src, dst, dtype, saturate)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
Reads an image, applies augmentations and writes results.
"""

import argparse
import os
import sys

import pyecvl.ecvl as ecvl


def main(args):
    if not ecvl.ECVL_EDDL:
        print("No EDDL support - quitting")
        sys.exit(0)
    head, ext = os.path.splitext(os.path.basename(args.in_fn))
    img = ecvl.ImRead(args.in_fn)
    c = ecvl.SequentialAugmentationContainer([ecvl.AugFlip(0.5)])
    da = ecvl.DatasetAugmentations([c, None, None])
    # some output images should be flipped
    for i in range(10):
        da.Apply(ecvl.SplitType.training, img)
        out_fn = "%s_flip_%d%s" % (head, i, ext)
        print("writing", out_fn)
        ecvl.ImWrite(out_fn, img)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("in_fn", metavar="INPUT_PATH")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
Compute image moments and draw an ellipse on the image.
"""

import argparse
import os
import sys

import numpy as np
import pyecvl.ecvl as ecvl


def main(args):
    head, ext = os.path.splitext(os.path.basename(args.in_fn))
    img = ecvl.ImRead(args.in_fn)
    origin = img.copy()
    ecvl.ChangeColorSpace(img, img, ecvl.ColorType.GRAY)
    ecvl.Threshold(img, img, 25, 1)
    print("Calculating moments")
    moments = ecvl.Image.empty()
    ecvl.Moments(img, moments, 2)

    # use moments to calculate mass/area and center of mass of the image
    rm = np.array(moments, copy=False)
    print("Raw moments:")
    print(rm)
    M00 = rm[0, 0]
    M10 = rm[1, 0]
    M01 = rm[0, 1]
    # M11 = rm[1, 1]
    # M02 = rm[0, 2]
    # M20 = rm[2, 0]

    # centroid coordinates
    x = M10 / M00
    y = M01 / M00

    # central moments
    ecvl.CentralMoments(img, moments, [x, y])
    cm = np.array(moments, copy=False)
    print("Central moments:")
    print(cm)
    u00 = cm[0, 0]
    u11 = cm[1, 1]
    u20 = cm[2, 0]
    u02 = cm[0, 2]

    # note that central moments can also be computed through the raw moments
    # u00_bis = M00
    # u11_bis = M11 - x * M01
    # u20_bis = M20 - x * M10
    # u02_bis = M02 - y * M01

    # terms of the covariance matrix
    u_20 = u20 / u00  # or M20 / M00 - (x * x)
    u_02 = u02 / u00  # or M02 / M00 - (y * y)
    u_11 = u11 / u00  # or M11 / M00 - (x * y)

    # Eigenvalues (lambda1, lambda2) and orientation of the eigenvector
    lambda1 = (u_20 + u_02) / 2 + np.sqrt(
        4 * u_11 * u_11 + (u_20 - u_02) * (u_20 - u_02)
    ) / 2
    lambda2 = (u_20 + u_02) / 2 - np.sqrt(
        4 * u_11 * u_11 + (u_20 - u_02) * (u_20 - u_02)
    ) / 2
    theta = 0.5 * np.arctan2(2 * u_11, (u_20 - u_02))  # rad

    # Eigenvalues are proportional to the square length of the eigenvector
    # axes. So the half-axes of the ellipse generated by the eigenvectors are
    # given by d/sqrt(lambda1) and d/sqrt(lambda2) where d is the proportional
    # factor. Considering that the moment M00 is the area (the image on which
    # it is calculated is binary) of the image objects we can calculate d.

    d = np.sqrt(M00 * np.sqrt(lambda1 * lambda2) / np.pi)

    # half-axes (a and b) are then
    a = d / np.sqrt(lambda1)
    b = d / np.sqrt(lambda2)

    # We can now draw the ellipses with the same moments as the image objects
    ecvl.DrawEllipse(origin, [int(x), int(y)], [int(a), int(b)],
                     theta * 180 / np.pi, [0, 0, 255], 2)

    out_path = "%s_moments%s" % (head, ext)
    print("Writing %s" % out_path)
    ecvl.ImWrite(out_path, origin)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("in_fn", metavar="INPUT_PATH")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import _core
_ecvl = _core.ecvl

ECVL_EDDL = _ecvl.ECVL_EDDL
ECVL_WITH_OPENSLIDE = _ecvl.ECVL_WITH_OPENSLIDE
ECVL_WITH_DICOM = _ecvl.ECVL_WITH_DICOM

from .arithmetic import *  # noqa: E402,F401,F403
from .dataset_generator import *  # noqa: E402,F401,F403
from .dataset_parser import *  # noqa: E402,F401,F403
from .datatype import *  # noqa: E402,F401,F403
from .image import *  # noqa: E402,F401,F403
from .imgproc import *  # noqa: E402,F401,F403
from .support_imgcodecs import *  # noqa: E402,F401,F403
from .support_nifti import *  # noqa: E402,F401,F403
if ECVL_EDDL:
    from .augmentations import *  # noqa: E402,F401,F403
    from .eddl_support import *  # noqa: E402,F401,F403
if ECVL_WITH_OPENSLIDE:
    from .support_openslide import *  # noqa: E402,F401,F403
if ECVL_WITH_DICOM:
    from .support_dcmtk import *  # noqa: E402,F401,F403



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import _core
_ecvl = _core.ecvl


__all__ = [
    "DataType",
    "DataTypeSize",
    "DataTypeSignedSize",
]


class DataType(_ecvl.DataType):
    """\
    Enum class which defines data types allowed for images.
    """
    int8 = _ecvl.DataType.int8
    int16 = _ecvl.DataType.int16
    int32 = _ecvl.DataType.int32
    int64 = _ecvl.DataType.int64
    float32 = _ecvl.DataType.float32
    float64 = _ecvl.DataType.float64
    uint8 = _ecvl.DataType.uint8
    uint16 = _ecvl.DataType.uint16
    none = _ecvl.DataType.none


def DataTypeSize(dt=None):
    """\
    Get the size in bytes of a given DataType.

    With no arguments, get the number of existing DataType members.

    :param dt: a DataType.
    :return: the DataType size in bytes, or the number of existing DataType
      members if called with no arguments
    """
    if dt is None:
        return _ecvl.DataTypeSize()
    return _ecvl.DataTypeSize(dt)


def DataTypeSignedSize():
    """\
    Get the number of existing signed DataType members.

    :return: the number of existing signed DataType members
    """
    return _ecvl.DataTypeSignedSize()



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import io
import os
import pytest

import pyecvl._core.ecvl as ecvl_core
import pyecvl.ecvl as ecvl_py


DATASET = """\
name: Foo
classes:
- '0'
- '1'
- '2'
images:
- location: foo_0.png
  label: 2
- location: foo_1.png
  label: 0
- location: foo_2.png
  label: 1
split:
  training:
  - 0
  - 1
  test:
  - 2
"""


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_load(ecvl, tmp_path):
    fn = str(tmp_path / "foo.yml")
    with io.open(fn, "wt") as f:
        f.write(DATASET)
    d = ecvl.Dataset(fn)
    assert d.name_ == "Foo"
    assert d.classes_ == ["0", "1", "2"]
    assert len(d.samples_) == 3
    with pytest.raises(RuntimeError):
        d.samples_[0].LoadImage()
    locations = [os.path.basename(_.location_[0]) for _ in d.samples_]
    assert locations == ["foo_0.png", "foo_1.png", "foo_2.png"]


@pytest.mark.parametrize("ecvl", [ecvl_core, ecvl_py])
def test_load_nonexistent(ecvl):
    with pytest.raises(RuntimeError):
        ecvl.Dataset("/not/existing.yml")
